from imagekit.processors import ResizeToFill, ResizeToFit

from pilkit.processors.resize import ResizeCanvas


class BaseInstanceProcessor(object):


    def process(self, image):
        if not self.image_instance:
            print("WARNING: Position crop expects image_instance, but none set.")
            return image

        return self.process_instance(image, self.image_instance)

    def process_instance(self, image, instance):
        raise NotImplementedError

    def get_hash(self):
        return self



class PositionCrop(BaseInstanceProcessor):
    """
    Processor to create custom image crops. Receieves image_instance and implemen

    """



    def __init__(self, options):
        self.options = options
        self.crop_position_field = options['crop_field']


        self.resize_method = options['resize_method']
        self.width = options['width']
        self.height = options['height']
        self.upscale = options['upscale'] or False

    def get_hash(self):
        #Hash based on crop value
        crop_value = getattr(self.image_instance, self.crop_position_field)
        hashed = u"%s-%s-%s-%s"%(self.crop_position_field, crop_value, self.width, self.height)
        # print hashed
        return hashed

    def get_crop_value(self, instance):
        return getattr(instance, self.crop_position_field)

    def process_instance(self, image, instance): 


        #Step 1, crop based on crop position
        crop_value = self.get_crop_value(instance)
        original_width = image.size[0]
        original_height = image.size[1]
        
        # print 'process %s :: original width %s original height %s || SETTINGS %s - %s - %s upscale? %s crop? %s, %s, %s, %s'%(instance, original_width, original_height, self.width, self.height, self.resize_method, self.upscale, crop_value.x, crop_value.y, crop_value.width, crop_value.height)
        is_empty_cropper = crop_value == None or crop_value == '' or (crop_value.width==None and crop_value.height==None)
        if is_empty_cropper:
            #Set crop with inital crop.

            # print 'crop: %s, %s original: %s, %s'%(self.width, self.height, original_width, original_height)

            crop_w = self.width#width if self.width is None else self.width
            if self.height and not crop_w:
                crop_w = int(float(float(self.height) / float(original_height))*float(original_width))



            crop_h = self.height#height if self.height is None else self.height
            if self.width and not crop_h:
                crop_h = int(float(float(self.width) / float(original_width))*float(original_height))

            if self.resize_method == 'fit':
                # print "Resize to fit: %s, %s"%(crop_w, crop_h)
                resizer = ResizeToFit(crop_w, crop_h, None, self.upscale)            
            else:
                # print "Resize to fill: %s, %s"%(crop_w, crop_h)
                resizer = ResizeToFill(crop_w, crop_h, None, self.upscale)

            resized = resizer.process(image)
            return resized
            
            
        crop_x = 0 if crop_value.x is None else int(0-crop_value.x)
        crop_y = 0 if crop_value.y is None else int(0-crop_value.y)
        crop_w = None if crop_value.width is None else int(crop_value.width)
        crop_h = None if crop_value.height is None else int(crop_value.height)

        # print "Resize canvas: %s, %s, %s, %s"%(crop_x, crop_y, crop_w, crop_h)
        cropper = ResizeCanvas(crop_w, crop_h, None, None, crop_x, crop_y)
        cropped = cropper.process(image)
       

        

        #Step 2, resize to correct width and height:
        if self.resize_method == 'fit':
            width = None if not self.width else self.width
            height = None if not self.height else self.height
            if height==None:
                # print 'height = (%s/%s) * %s'%(width, original_width, original_height)
                height = int(float(float(width)/float(original_width)) * float(original_height))
            elif width==None:
                width = int(float(float(height)/float(original_height)) * float(original_width))


            # print "Resize to fit: %s, %s"%(width, height)
            resizer = ResizeToFit(width, height, None, self.upscale)            
        else:
            width = cropped.size[0] if not self.width else self.width
            height = cropped.size[1] if not self.height else self.height
            # print "Resize to fill: %s, %s"%(crop_w, crop_h)
            resizer = ResizeToFill(width, height, None, self.upscale)
        
        # print 'Resize to %s - %s (%s - %s)'%(width, height, resizer.width, resizer.height)
        resized = resizer.process(cropped)
        return resized



class PositionAndFormatCrop(PositionCrop):
    """
    Resize based on crop and format

    """

    def __init__(self, options):
        self.options = options
        
        self.crop_position_field = options['crop_field']
        self.format_field = options['format_field']
        self.resize_method = options['resize_method']
        self.width = options['width']
        self.height = options['height']
        self.upscale = options['upscale'] or False

    def get_hash(self):
        #Hash based on crop value
        crop_value = getattr(self.image_instance, self.crop_position_field)
        format = getattr(self.image_instance, self.format_field)
        hashed = u"%s-%s-%s-%s-%s"%(self.crop_position_field, crop_value, format, self.width, self.height)
        # print hashed
        return hashed

class FormatProcessor(PositionCrop):
    """
    Resize based on format; use default crop

    """

    def __init__(self, options):
        self.options = options
        self.format_field = options['format_field']
        self.resize_method = options['resize_method']
        self.width = options['width']
        self.height = options['height']
        self.upscale = options['upscale'] or False

    def get_crop_value(self, instance):
        return None

    def get_hash(self):
        #Hash based on crop value
        format = getattr(self.image_instance, self.format_field)
        hashed = u"%s-%s-%s"%(format, self.width, self.height)
        # print hashed

        return hashed        




from django.contrib.admin import widgets
from django.contrib.admin.utils import help_text_for_field
from django.utils.safestring import mark_safe

class ImageCropWidget(widgets.AdminTextInputWidget):
	image_model = None

	class Media:
		css = {
			'all': ('imagekit_cropper/jcrop/css/jquery.Jcrop.css', 'imagekit_cropper/css/imagecrop.css',)
		}
		js = ('imagekit_cropper/js/vendor/jquery.js','imagekit_cropper/js/imagecrop.js','imagekit_cropper/jcrop/js/jquery.Jcrop.js')
	
	def __init__(self, properties, help_text='', *args, **kwargs):
		self.properties = properties
		self.help_text = help_text
		
		return super(ImageCropWidget, self).__init__(*args, **kwargs)

		

	def get_crop_containers(self):
		properties = self.properties
		if 'aspect_ratio' in self.properties:
			return '<div class="image-cropper" data-width="%s" data-height="%s" data-ratio="%s" \
			data-resize-method="%s" data-source="%s" data-upscale="%s" ></div>\
			'%(self.properties['min_width'], self.properties['min_height'],
			self.properties['aspect_ratio'], self.properties['resize_method'],
			self.properties['source'], self.properties['upscale'])

		else:
			return '<div class="image-cropper" data-width="%s" data-height="%s" \
				data-resize-method="%s" data-source="%s" data-upscale="%s" ></div>\
				'%(self.properties['width'], 
				self.properties['height'], self.properties['resize_method'],
				self.properties['source'], self.properties['upscale'])

	def render(self, name, value, attrs=None):
		rendered = super(ImageCropWidget, self).render(name, value, attrs)
		return mark_safe('<div class="image-crop-container">%s%s<div \
			class="image-crop-data"><p class="grp-help">%s</p></div>\
			</div>'%(rendered, self.get_crop_containers(), self.help_text))


from django.db import models

from imagekit.models import ImageSpecField
from imagekit.models.fields.utils import ImageSpecFileDescriptor
from imagekit.registry import generator_registry, unregister
from imagekit.specs import SpecHost, create_spec_class
from imagekit.specs.sourcegroups import ImageFieldSourceGroup

from .specs import InstanceSpec, InstanceFormatSpec
from .registry import InstanceSourceGroupRegistry, instance_source_group_registry


class InstanceSpecField(ImageSpecField):
    """
    Image spec field which passes the class instance to the processors

    """


    fields = None
    source = None

    def __init__(self, processors=None, format=None, options=None,
            source=None, cachefile_storage=None, autoconvert=None,
            cachefile_backend=None, cachefile_strategy=None, spec=None,
            id=None):


        self.options = options
        self.format = format
        self.processors = processors

        
        #==== FROM ImageSpecField
        SpecHost.__init__(self, processors=processors, format=format,
                options=options, cachefile_storage=cachefile_storage,
                autoconvert=autoconvert,
                cachefile_backend=cachefile_backend,
                cachefile_strategy=cachefile_strategy, spec=spec,
                spec_id=id)

        # TODO: Allow callable for source. See https://github.com/matthewwithanm/django-imagekit/issues/158#issuecomment-10921664
        self.source = source
        #==== END FROM ImageSpecField

        self.spec = spec = self.get_spec_class() 


        data = self.get_spec_instance_attrs()
        for attr in data:
            setattr(self.spec, attr, data[attr])

        spec_id = id

        self._original_spec = spec

        #==== END FROM SpecHost

    def contribute_to_class(self, cls, name):
        
        descriptor = ImageSpecFileDescriptor(self, name, self.source)
        setattr(cls, name, descriptor)

        #Store spec instance info based on spec ID
        spec_id = ('%s:%s:%s' % (cls._meta.app_label,
                            cls._meta.object_name, name)).lower()
        
        #Un-register with default source group; use custom
        unregister.source_group(spec_id, ImageFieldSourceGroup(cls, self.source))
        instance_source_group_registry.register(spec_id, ImageFieldSourceGroup(cls, self.source), self.get_spec_instance_attrs())  

        # setattr(cls, name, InstanceSpecFileDescriptor(self, name, source))
        self._set_spec_id(cls, name)
        
    def get_spec(self, source):
        
        item = generator_registry.get(self.spec_id, source=source, specs=self.get_spec_instance_attrs())

        return item


    def get_spec_class(self):
        return InstanceSpec

    def get_spec_instance_attrs(self):
        return {'options': self.options, 'processors': self.processors, 'format': self.format} #override


class InstanceFormatSpecField(InstanceSpecField):

    def __init__(self, format_field, processors=None, format=None, options=None,
            source=None, cachefile_storage=None, autoconvert=None,
            cachefile_backend=None, cachefile_strategy=None, spec=None,
            id=None):
        
        self.format_field = format_field

        super(InstanceFormatSpecField, self).__init__(processors, format, 
            options, source, cachefile_storage, autoconvert, cachefile_backend, 
            cachefile_strategy, spec, id)

    def get_spec_instance_attrs(self):
        return {'options': self.options, 'processors': self.processors, 'format': self.format, 'format_field':self.format_field} #override

    def get_spec_class(self):
        return InstanceFormatSpec

class CropCoordinates(object):
    def __init__(self, x=None, y=None, width=None, height=None):
        self.x = x
        self.y = y
        self.width = width
        self.height = height

    def __repr__(self):
        if self.width or self.height:
            return "%s,%s,%s,%s"%(self.x, self.y, self.width, self.height)
        return ''

class ImageCropField(models.Field):
    """
    Model field for containing image crop dimensions

    """

    description = "Image crop coordinates"
    
    def __init__(self,properties, help_text=("A comma-separated list of crop coordinates"),verbose_name='imagecropfield', *args,**kwargs):
        self.name="ImageCropField",
        self.through = None
        self.help_text = help_text
        self.blank = True
        self.editable = True
        self.creates_table = False
        self.db_column = None
        self.serialize = False
        self.null = True
        self.creation_counter = models.Field.creation_counter        
        self.properties = properties
        # self.default_width = self.properties['width'] if self.properties['width'] else 1000
        # self.default_height = self.properties['height'] if self.properties['height'] else 1000
        models.Field.creation_counter += 1
        super(ImageCropField, self).__init__(*args, **kwargs)


    def deconstruct(self):
        name, path, args, kwargs = super(ImageCropField, self).deconstruct()
        kwargs['properties'] = self.properties
        return name, path, args, kwargs
        
        
    def db_type(self, connection):
        return 'varchar(100)'
    
    def to_python(self,value):
        
        if value in ( None,''):
            
            return CropCoordinates()
        else:
            if isinstance(value, CropCoordinates):
                return value
            else:

                if 'None' in value:
                    return CropCoordinates()

                split_items = value.split(',')

                

                x = float(split_items[0])
                y = float(split_items[1])
                w = float(split_items[2])
                h = float(split_items[3])

                args = [x,y,w,h]
                if len(args) != 4 and value is not None:
                    raise ValidationError("Invalid input for a CropCoordinates instance")
                return CropCoordinates(*args)
         

    def get_prep_value(self, value):
        if value:
            store_value = ','.join([str(value.x),str(value.y),str(value.width),str(value.height)])
            return store_value
        return None
    
    def get_internal_type(self):
        return 'CharField'
    
    def value_to_string(self, obj):
        value = self._get_val_from_obj(obj)
        return self.get_prep_value(value)    



import re
import traceback

from django.core.management.base import BaseCommand

from imagekit.registry import generator_registry, cachefile_registry
from imagekit.exceptions import MissingSource
from imagekit.utils import  call_strategy_method
from imagekit.cachefiles import ImageCacheFile

from imagekit_cropper.registry import spec_data_field_hash


class Command(BaseCommand):
    help = ("""Generate files for the specified image generators (or all of them if
none was provided). Simple, glob-like wildcards are allowed, with *
matching all characters within a segment, and ** matching across
segments. (Segments are separated with colons.) So, for example,
"a:*:c" will match "a:b:c", but not "a:b:x:c", whereas "a:**:c" will
match both. Subsegments are always matched, so "a" will match "a" as
well as "a:b" and "a:b:c".""")
    args = '[generator_ids]'

    def handle(self, *args, **options):
        generators = generator_registry.get_ids()

        if args:
            patterns = self.compile_patterns(args)
            generators = (id for id in generators if any(p.match(id) for p in patterns))

        for generator_id in generators:
            self.stdout.write('Validating generator: %s\n' % generator_id)


            try:
                for image_file in cachefile_registry.get(generator_id):
                    if image_file.name:
                        self.stdout.write('  %s\n' % image_file.name)
                        try:
                            image_file.generate()
                        except Exception as err:
                            self.stdout.write('\tFailed %s\n' % (err))
            except:
                items = cachefile_registry._cachefiles.items()
                
                for k, v in items:

                    
                    if generator_id in v:
                        #k is SourceGroupFilesGenerator
                       

                        model_class = k.source_group.model_class
                        image_field = k.source_group.image_field
                        all_objects = model_class.objects.all()
                        item_count = len(all_objects)
                        count = 0

                        for instance in all_objects:
                            count += 1

                            print 
                            try:
                                source = getattr(instance, image_field)
                                specs = spec_data_field_hash[generator_id]

                                spec = generator_registry.get(generator_id, source=source, specs=specs)

                                file = ImageCacheFile(spec)
                                self.stdout.write('  [%s of %s] - %s\n' % (count, item_count, file))
                                call_strategy_method(file, 'on_source_saved')                         

                            except Exception:

                                self.stdout.write('ERROR: %s\n' % (traceback.format_exc()))
                        

    def compile_patterns(self, generator_ids):
        return [self.compile_pattern(id) for id in generator_ids]

    def compile_pattern(self, generator_id):
        parts = re.split(r'(\*{1,2})', generator_id)
        pattern = ''
        for part in parts:
            if part == '*':
                pattern += '[^:]*'
            elif part == '**':
                pattern += '.*'
            else:
                pattern += re.escape(part)
        return re.compile('^%s(:.*)?$' % pattern)
